__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "vivarium_inputs"
__summary__ = "Transformations and artifact building for the vivarium microsimulation project."
__uri__ = "https://github.com/ihmeuw/vivarium_inputs"

__version__ = "1.0.9"

__author__ = "The vivarium developers"
__email__ = "vivarium.dev@gmail.com"

__license__ = "GNU GPLv3"
__copyright__ = f"Copyright 2016-2018 {__author__}"
